"""Example Utility Functions."""


def multiply(first_digit: float, second_digit: float) -> float:
    """Multiplies two floats and returns the result."""
    result = first_digit * second_digit
    return result


def divide(first_digit: float, second_digit: float) -> float:
    """Divides first float by second one and returns the result."""
    result = first_digit / second_digit
    return result



"""Create a fully implemented sql connection engine."""

import os
import struct

from azure.identity import DefaultAzureCredential
from sqlalchemy import Engine, create_engine
from sqlalchemy.engine import URL


def create_connection_engine() -> Engine:
    """Create a fully implemented sql connection engine."""
    credentials = os.getenv("URL_AZURECLICREDENTIALTOKEN")
    tsql_driver = os.getenv("TRANSACTSQLDRIVER")
    tsql_server = os.getenv("TRANSACTSQLSERVER")
    tsql_database = os.getenv("TRANSACTSQLDATABASE")
    tsql_access_token = int(os.getenv("TRANSACTSQLACCESSTOKEN"))

    credential = DefaultAzureCredential()
    databaseToken = credential.get_token(credentials)
    token_bytes = bytes(databaseToken[0], "UTF-8")
    exptoken = b"\x00".join(bytes({i}) for i in token_bytes) + b"\x00"
    tokenstruct = struct.pack("=i", len(exptoken)) + exptoken

    connString = f"Driver={tsql_driver};Server={tsql_server};Database={tsql_database}"
    args_con = {"attrs_before": {tsql_access_token: tokenstruct}}
    connUrl = URL.create("mssql+pyodbc", query={"odbc_connect": connString})

    engine = create_engine(connUrl, connect_args=args_con)
    return engine



"""This file contains an example of an ETL-operation on the Bag AO dataset."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from sqlalchemy import Connection

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def bag_ao_etl(sql_connection: Connection) -> None:
    """Execute an ETL-operation on the BAG AO dataset.

    Uses a generic sql-connection.
    """
    dataset = extract(sql_connection=sql_connection)
    dataset = transform(dataset)
    load(dataset)


def extract(sql_connection: Connection) -> pd.DataFrame:
    """Extract the BAG AO dataset."""
    sql = "SELECT * FROM [Current].[LVBAG2_0_ADRES]"
    dtpye_dict = {
        "openbare_ruimte_id": str,
        "nummeraanduiding_id": str,
        "woonplaats_id": str,
        "naam_van_openbare_ruimte": str,
        "huisnummer": float,
        "postcode": str,
        "huisletter": str,
        "huisnummertoevoeging": str,
        "begin_geldigheid": str,
        "eind_geldigheid": str,
        "current_indicator": int,
    }
    dataset = pd.DataFrame()
    for connection_attempt in range(1, 6):
        try:
            dataset = pd.read_sql(sql=sql, con=sql_connection, dtype=dtpye_dict)

            if dataset is None or len(dataset) == 0:
                continue

            return dataset

        except Exception as error:
            errormsg = f"An error occured during attempt {connection_attempt}: {error!r}"
            logger.warning(errormsg)
            continue

    errormsg = f"Failed to load data rows for {sql}"
    logger.error(errormsg)
    raise Exception(errormsg)


def transform(df: pd.DataFrame) -> pd.DataFrame:
    """Transform the dataset."""
    # Verwijder duplicate nummers
    df = df.drop_duplicates(subset="nummeraanduiding_id", keep="last")

    # Geldigheid altijd in de toekomst
    now = pd.Timestamp("today")
    df["begin_geldigheid"] = pd.to_datetime(df["begin_geldigheid"])
    df["eind_geldigheid"] = pd.to_datetime(df["eind_geldigheid"], errors="coerce")
    mask = df["begin_geldigheid"].notna()
    mask &= df["begin_geldigheid"] < now
    mask &= df["eind_geldigheid"].isna() | (df["eind_geldigheid"] > now)
    df = df[mask]

    # Geldig huisnummer
    df = df[df["huisnummer"].notna()]
    df = df.astype({"huisnummer": int})

    return df


def load(df: pd.DataFrame) -> None:
    """Load the dataset to the target.

    Throws an exception if the dataset is empty.
    """
    if len(df) == 0:
        errormsg = "Mayday! We have found an empty database."
        raise Exception(errormsg)
    logger.info("Dataframe is now ready to be loaded!")


if __name__ == "__main__":
    # Do not import this from the head of the file
    # A reference to a production sql-connection(/engine) should not be made unless necesarry
    # For example, unit tests will fail when this import is made.
    # Ideally, such a reference should be imported from within the project_repo application
    # For sake of the example, this will suffice.
    from dotenv import find_dotenv, load_dotenv
    from example_sql_connection import create_connection_engine

    load_dotenv(find_dotenv())

    sql_engine = create_connection_engine()

    with sql_engine.connect() as sql_connection:
        bag_ao_etl(sql_connection=sql_connection)



"""Logger module."""
# Copyright Gemeente Rotterdam - All Rights Reserved
# Unauthorized copying of this file, via any medium is strictly prohibited
# Proprietary and confidential

import logging
import sys

log_format = "%(levelname)s %(asctime)s - %(message)s"
logging.basicConfig(stream=sys.stdout, filemode="w", format=log_format, level=logging.INFO)

logger = logging.getLogger("azure")
logger.setLevel(logging.INFO)
